library nui_push;

import 'package:flutter/cupertino.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/code/nui_device_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/model/interfaces.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_push/model/nui_push_callback.dart';
import 'package:nui_push/model/nui_push_models.dart';
import 'package:nui_web/nui_web.dart';
import 'package:onesignal_flutter/onesignal_flutter.dart';

export 'package:nui_push/model/nui_push_callback.dart';
export 'package:nui_push/model/nui_push_models.dart';

abstract class NUIPushInterface{
  Future<DataResult> registerPushDeviceForUser({@required String userId});
  Future<DataResult> updateReadStatus({@required String recipientId});
  Future<DataResult> setPushDeviceMeta({@required String userId, @required Map<String, dynamic> params});
  Future setLocationShared(bool enable);
  Future<DataResult> deletePushDeviceMeta({@required String userId, @required List<String> params});
  Future<DataResult<List<NUIPushMessage>>> getPushMessages({int page, int limit});
  Future<DataResult<NUIPushMessage>> getPushMessageById({@required String recipientId});
  Future<Map<String, dynamic>> getTags();
  Future sendTag(String key, dynamic value);
  Future<Map<String, dynamic>> sendTags(Map<String, dynamic> tags);
  Future deleteTag(String key);
  Future deleteTags(List<String> keys);
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToUsers({String title, String subtitle, NUIEnt data, List<String> userIds, String bigPictureUrl});
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToAll({String title, String subtitle, NUIEnt data, List<String> includedSegments, List<String> excludedSegments, String bigPictureUrl});
}

class NUIPush implements NUIPushInterface{
  static const String _MODULE = "NUIPush";
  static const String _NUIWEB_MODULE = "NUIPushWebModule";
  static const String _NUIWEB_ONESIGNAL = "NUIPushWebOSModule";
  static const String _ONE_SIGNAL_REST_URL = "https://onesignal.com/api/v1/notifications";
  static const String _OS_USER_ID = "NUIPushUserID";

  static NUIPush _instance;
  static NUIPush get instance {
    if(_instance == null){
      throwNUI("NUIPush", "NUIPush has not been built with NUIPushBuilder");
    }
    return _instance;
  }

  NUIPush._builder(NUIPushBuilder builder){
    _baseUrl = builder._baseUrl;
    _oneSignalAuthKey = builder._oneSignalAuthKey;
    _oneSignalAppId = builder._oneSignalAppId;
    _notificationReceivedListener = builder._notificationReceivedListener;
    _notificationOpenedListener = builder._notificationOpenedListener;
    _instance = this;
  }

  String _baseUrl;
  String _oneSignalAuthKey;
  String _oneSignalAppId;
  NUIPushNotificationListener _notificationReceivedListener;
  NUIPushNotificationListener _notificationOpenedListener;

  @override
  Future<DataResult<NUIPushMessage>> getPushMessageById({String recipientId}) async{
    final result = await NUIWeb.get(module: _NUIWEB_MODULE).doGET<NUIPushMessage>(url: "push/my-messages", pathVariable: recipientId, entity: NUIPushMessageEntity());
    if(result.success == true){
      logNUI(_MODULE, "Get push message of recipient id : $recipientId successful");
      return DataResult<NUIPushMessage>(success: true, message: result.message, data: result.body);
    }
    else{
      logNUI(_MODULE, "Failed to get push message of recipient id : $recipientId");
      return DataResult(success: false, message: result.message);
    }
  }

  void _saveUserID(String userId) async{
    await NUISharedPref.get().save(_OS_USER_ID, userId);
  }

  Future<String> getUserID() async{
    return NUISharedPref.get().getString(_OS_USER_ID, defValue: "");
  }

  void _registerUserIDIfChanged(String userId) async{
    final prevUserId = await NUISharedPref.get().getString(_OS_USER_ID);
    if(!match(prevUserId, userId)){
      registerPushDeviceForUser(userId: userId);
    }
  }

  @override
  Future setLocationShared(bool enable) async {
    final hasPermission = await NUIPermissionUtil.hasLocationPermission();
    if(hasPermission) {
      final result = await OneSignal.shared.setLocationShared(enable);
    }
  }

  @override
  Future<DataResult<List<NUIPushMessage>>> getPushMessages({int page, int limit}) async{
    final result = await NUIWeb.get(module: _NUIWEB_MODULE).doGET(url: "push/my-messages", params: {
      "page": page?.toString(),
      "limit": limit?.toString()
    });

    if(result.success == true){
      logNUI(_MODULE, "Get push messages successful");
      return DataResult<List<NUIPushMessage>>(success: true, message: result.message, data: asListOfObjects(NUIPushMessageEntity(), result.responseBody));
    }
    else{
      logNUI(_MODULE, "Failed to get push messages");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult> registerPushDeviceForUser({String userId}) async{
    if(isNullOrEmpty(userId)) return DataResult(success: false, message: "Failed to register device due to missing device token");
    OneSignal.shared.setSubscriptionObserver((changes) async{
      final osStatus = changes.to ?? changes.from;
      final deviceToken = osStatus.userId;

      final deviceInfo = await NUIDeviceUtil.getDeviceInfo();
      final registerRequest = NUIRegisterPushDeviceRequest(userId: userId, uniqueId: deviceInfo.uuid ?? "", providerDeviceToken: deviceToken, deviceOs: deviceInfo.platform, deviceModel: deviceInfo.model);

      final result = await NUIWeb.get(module: _NUIWEB_MODULE).doPOST<NUIStandardHttpResponse>(url: "push/registerNew", bodyEnt: registerRequest, entity: NUIStandardHttpResponseEntity());
      if(result.success == true){
        logNUI(_MODULE, "Register device successful");
        return DataResult(success: true, message: "Register device successful");
      }
      else{
        logNUI(_MODULE, "Failed to register device");
        return DataResult(success: false, message: result.message);
      }
    });
    return DataResult(success: true, message: "Device will be registered upon receiving new OS ID");
  }

  @override
  Future<DataResult> setPushDeviceMeta({String userId, Map<String, dynamic> params}) async{
    if(isNullOrEmpty(userId)) return DataResult(success: false, message: "Invalid user ID: $userId");
    final result = await NUIWeb.get(module: _NUIWEB_MODULE).doPOST<NUIStandardHttpResponse>(url: "push/devices/$userId/user-meta", body: params, entity: NUIStandardHttpResponseEntity());
    final updateResult = await sendTags(params);
    if(result.success == true){
      logNUI(_MODULE, "Update device push metadata successful");
      return DataResult(success: true, message: "Update device push metadata successful");
    }
    else{
      logNUI(_MODULE, "Failed to update device push metadata");
      return DataResult(success: false, message: result.message);
    }
  }

  Future<Map<String, dynamic>> getTags() async{
    return OneSignal.shared.getTags();
  }

  Future sendTag(String key, dynamic value) async{
    await OneSignal.shared.sendTag(key, value);
  }

  Future<Map<String, dynamic>> sendTags(Map<String, dynamic> tags) async{
    return OneSignal.shared.sendTags(tags);
  }

  Future deleteTag(String key) async{
    await OneSignal.shared.deleteTag(key);
  }

  Future deleteTags(List<String> keys) async{
    await OneSignal.shared.deleteTags(keys);
  }

  @override
  Future<DataResult> deletePushDeviceMeta({@required String userId, @required List<String> params}) async{
    if(isNullOrEmpty(userId)) return DataResult(success: false, message: "Invalid user ID: $userId");
    final result = deleteTags(params);
    logNUI(_MODULE, "Delete device push metadatas successful");
    return DataResult(success: true, message: "Update device push metadata successful");
  }

  @override
  Future<DataResult> updateReadStatus({String recipientId}) async{
    final result = await NUIWeb.get(module: _NUIWEB_MODULE).doPUT<NUIStandardHttpResponse>(url: "push/message/read", pathVariable: recipientId, entity: NUIStandardHttpResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Update message read status successful");
      return DataResult(success: true, message: "Update message read status successful");
    }
    else{
      logNUI(_MODULE, "Failed to update message read status");
      return DataResult(success: false, message: result.message);
    }
  }

  @override
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToAll({String title, String subtitle, NUIEnt data, List<String> includedSegments, List<String> excludedSegments, String bigPictureUrl}) async{
    try{
      logNUI(_MODULE, "Sending push notification to all");
      final notification = NUIPushOSNotification(
          appId: _oneSignalAppId,
          headings: {"en": title},
          subtitle: {"en": subtitle},
          contents: {"en": subtitle},
          includedSegments: includedSegments,
          excludedSegments: excludedSegments,
          data: data.toMap(),
          bigPicture: bigPictureUrl
      );

      final result = await NUIWeb.get(module: _NUIWEB_ONESIGNAL).doPOST<NUIPushOSResponse>(bodyEnt: notification, headers: _oneSignalPushHeaders(), entity: NUIPushOSResponseEntity());
      if(result.success == true){
        logNUI(_MODULE, "Send push message to all successful");
        return DataResult(success: true, message: "Send push message to all successful");
      }
      else{
        logNUI(_MODULE, "Failed to send push message to all with message: ${result.message}");
        return DataResult(success: false, message: "Failed to send push message to all with message : ${result.message}");
      }
    }catch(e){
      logNUI(_MODULE, "Failed to send push message to all with error : $e");
      return DataResult(success: false, message: "Failed to send push message to all with error: $e");
    }
  }

  Map<String, String> _oneSignalPushHeaders(){
    return {
      "Content-Type": "application/json; charset=UTF-8",
      "Authorization": "Basic $_oneSignalAuthKey"
    };
  }

  @override
  Future<DataResult<NUIPushOSResponse>> sendPushMessageToUsers({String title, String subtitle, NUIEnt data, List<String> userIds, String bigPictureUrl}) async{
    try{
      logNUI(_MODULE, "Sending push notification to users : $userIds");
      final notification = NUIPushOSNotification(
        appId: _oneSignalAppId,
        headings: {"en": title},
        subtitle: {"en": subtitle},
        contents: {"en": subtitle},
        includePlayerIds: userIds,
        data: data.toMap(),
        bigPicture: bigPictureUrl
      );

      final result = await NUIWeb.get(module: _NUIWEB_ONESIGNAL).doPOST<NUIPushOSResponse>(bodyEnt: notification, headers: _oneSignalPushHeaders(), entity: NUIPushOSResponseEntity());
      if(result.success == true){
        logNUI(_MODULE, "Send push message to users successful");
        return DataResult(success: true, message: "Send push message to users successful");
      }
      else{
        logNUI(_MODULE, "Failed to send push message to users with message: ${result.message}");
        return DataResult(success: false, message: "Failed to send push message to users with message : ${result.message}");
      }
    }catch(e){
      logNUI(_MODULE, "Failed to send push message to users with error : $e");
      return DataResult(success: false, message: "Failed to send push message to users with error: $e");
    }
  }
}

class NUIPushBuilder{
  String _baseUrl;
  String _oneSignalAuthKey;
  String _oneSignalAppId;
  NUIPushNotificationListener _notificationReceivedListener;
  NUIPushNotificationListener _notificationOpenedListener;
  NUIPushUserIDListener _notificationUserIDListener;

  NUIPushBuilder baseUrl(String baseUrl){
    _baseUrl = baseUrl;
    return this;
  }

  NUIPushBuilder oneSignalAuthKey(String authKey){
    _oneSignalAuthKey = authKey;
    return this;
  }

  NUIPushBuilder userIDChangedListener(NUIPushUserIDListener listener){
    _notificationUserIDListener = listener;
    return this;
  }

  NUIPushBuilder oneSignalAppId(String appId){
    _oneSignalAppId = appId;
    return this;
  }

  NUIPushBuilder notificationReceivedListener(NUIPushNotificationListener listener){
    _notificationReceivedListener = listener;
    return this;
  }

  NUIPushBuilder notificationOpenedListener(NUIPushNotificationListener listener){
    _notificationOpenedListener = listener;
    return this;
  }

  Future<NUIPush> build() async{
    final sharedPref = await NUISharedPrefBuilder()
      .build();

    NUIWebBuilder()
      .module(NUIPush._NUIWEB_MODULE)
      .headers(() => {
        "Content-Type": "application/json"
      })
      .baseURL(_baseUrl)
      .build();
    NUIWebBuilder()
      .headers(() => {
        "Content-Type": "application/json"
      })
      .module(NUIPush._NUIWEB_ONESIGNAL)
      .baseURL(NUIPush._ONE_SIGNAL_REST_URL)
      .build();
    final push = NUIPush._builder(this);

    OneSignal.shared.setLogLevel(OSLogLevel.verbose, OSLogLevel.none);

    OneSignal.shared.setAppId(_oneSignalAppId);
    // OneSignal.shared.init(
    //     _oneSignalAppId,
    //     iOSSettings: {
    //       OSiOSSettings.autoPrompt: false,
    //       OSiOSSettings.inAppLaunchUrl: false
    //     }
    // );

    // OneSignal.shared.setInFocusDisplayType(OSNotificationDisplayType.notification);

    await OneSignal.shared.promptUserForPushNotificationPermission(fallbackToSettings: true);
    OneSignal.shared.setNotificationWillShowInForegroundHandler((OSNotificationReceivedEvent event) {
      final notification = event.notification;
      // will be called whenever a notification is received
      if(_notificationReceivedListener != null) _notificationReceivedListener(NUIPushNotification(osNotification: notification));
    });

    OneSignal.shared.setNotificationOpenedHandler((OSNotificationOpenedResult result) {
      // will be called whenever a notification is opened/button pressed.
      if(_notificationOpenedListener != null) _notificationOpenedListener(NUIPushNotification(osNotification: result.notification));
    });

    OneSignal.shared.setPermissionObserver((OSPermissionStateChanges changes) {
      // will be called whenever the permission changes
      // (ie. user taps Allow on the permission prompt in iOS)
    });

    OneSignal.shared.setSubscriptionObserver((OSSubscriptionStateChanges changes) {
      // will be called whenever the subscription changes
      //(ie. user gets registered with OneSignal and gets a user ID)
      if(!isNullOrEmpty(changes.to?.userId)){
        push._registerUserIDIfChanged(changes.to.userId);
        push._saveUserID(changes.to.userId);
        if(_notificationUserIDListener != null) _notificationUserIDListener(changes.to.userId);
      }
    });

    OneSignal.shared.setEmailSubscriptionObserver((OSEmailSubscriptionStateChanges emailChanges) {
      // will be called whenever then user's email subscription changes
      // (ie. OneSignal.setEmail(email) is called and the user gets registered
    });

    return push;
  }
}